package org.omegahat.DataStructures.Data;

import org.omegahat.Environment.Databases.Database;
import org.omegahat.Environment.Databases.ObjectDatabase;

import java.io.StreamTokenizer;

/**
  Represents a collection of {@link VariableInt VariableInt}
  objects as a named group for accessing them individually
  or in sub-groups,
  indexable by name, and reducable/subsettable by
   observation index or observation name. 
*/
public class DataFrame 
//extends org.omegahat.Environment.Databases.ObjectDatabase
extends org.omegahat.Environment.Utils.OrderedTable
       implements java.io.Serializable,
                  org.omegahat.DataStructures.Data.DataFrameInt,
                  org.omegahat.Environment.Tools.DataScanner.RecordStreamListener
{
static public VariableInt UnknownVariable = new RealVariable(0);

  /**
    The number of observations in this 
    dataset.
  */
  private int length;    // No. of observations


  /**
    An ordered collection of identifiers
    for the different records in the dataset.
  */
  protected String[] rowNames;

  protected boolean readHeader;

  private int numHeadersSet = 0;

  protected java.util.Hashtable metaData;

  /**
   Constructor that provides no information about the number or types
   of variables or the number of observations.
  */
  public DataFrame() {
   
  }

  /**
    Constructor specifying the number of records that will be in this 
    dataset.
  */
  public DataFrame(Integer l) {
    this(l.intValue());
  }
  /**
    Constructor specifying the number of records that will be in this 
    dataset.
  */
  public DataFrame(int l) {
    this(l, new ObjectDatabase());
  }

  /**
    Constructor specifying the number of records that will be in this 
    dataset and a database containing variables to which should
   be added to this dataset. Each element in the database
    is checked to see if it is a variable and whether its
    length matches the specified value.
  */
  public DataFrame(int l, Database dbase) {
    length = l;

    if(dbase.size() > 0) {
      String[] names = dbase.objects();
      for(int i = 0; i < names.length; i++) {
        Object o = dbase.get(names[i]);
        if(!(o instanceof VariableInt) || !(((VariableInt) o).numObservations() == l))
          throw new IllegalArgumentException("All database objects should be"
                                              + " Variables with length " +
                                              String.valueOf(l));

        try {
           add(names[i], (VariableInt)o);
        } catch(Exception e) {
        }
      }
    }
  }


  /**
   Reads the contents of the data from the specified scanner by being
   notified when a complete record has been consumed by the scanner.
  */
  public DataFrame(org.omegahat.Environment.Tools.DataScanner.EventDataScanner scanner) {
   this(scanner, true);
  }

  /**

  */
  public DataFrame(org.omegahat.Environment.Tools.DataScanner.EventDataScanner scanner, boolean readHeader) {
   read(scanner, readHeader);
  }

  /**

  */
  public DataFrame(java.io.Reader reader, boolean readHeader) {
   read(new org.omegahat.Environment.Tools.DataScanner.EventDataScanner(reader), readHeader);
  }

  /**
   Reads the values from the specified connection/stream.
   @see #read(org.omegahat.Environment.Tools.EventDataScanner,boolean)
  */
  public DataFrame(java.io.Reader reader) {
   read(new org.omegahat.Environment.Tools.DataScanner.EventDataScanner(reader), true);
  }


  /**
   Reads the values from the specified stream by converting it to a reader
   and then creating a {@link org.omegahat.Environment.Tools.EventDataScanner EventDataScanner}.
   @see #read(org.omegahat.Environment.Tools.DataScanner.EventDataScanner,boolean)
  */
  public DataFrame(java.io.InputStream stream) {
   this(stream, true);
  }

  /**
   @see #read(org.omegahat.Environment.Tools.EventDataScanner,boolean)
  */
  public DataFrame(java.io.InputStream stream, boolean readHeader) {
   this(new java.io.InputStreamReader(stream), readHeader);
  }

  public DataFrame(java.io.File file, boolean readHeader) 
    throws java.io.FileNotFoundException
  {
   this(new java.io.FileReader(file), readHeader);
  }

  public DataFrame(java.io.File file) 
    throws java.io.FileNotFoundException
  {
   this(file, true);
  }



  public String[] getVariableNames() {
    return objects();
  } 

  /*
  public Variable get(String name) {
    return((Variable) db.get(name));
  }

  public void remove(String name) {
    db.remove(name);
  }
  */

  public void add(String name, VariableInt v) throws Exception {
    if(v.numObservations() == length || length == 0) {
      assign(name, v);
      if(length == 0)
         length = (int) v.numObservations();
    } else throw new Exception("Variable length should be same as "
                               + String.valueOf(length));
  }


  public Object[] get(int obsNo) {
    String[] names = objects();
    Object[] objs = new Number[names.length];
    for(int i = 0; i < names.length; i++) {
      objs[i] = ((VariableInt) get(names[i])).value(obsNo);
    }
    return(objs);
  }

  public DataFrame get(String[] name) throws Exception {
    DataFrame dframe = new DataFrame(length);
    for(int i = 0; i < name.length; i++) {
      VariableInt var = (VariableInt) get(name[i]);
      if(var != null)
        dframe.add(name[i], var);
    }
    return(dframe);  
  }

  public DataFrame get(int[] obsNos) throws Exception {
    String[] names = objects();
    DataFrame dframe = new DataFrame(obsNos.length);
    for(int i = 0; i < names.length; i++) {
      dframe.add(names[i], ((VariableInt) get(names[i])).values(obsNos));
    }
    return(dframe);
  }

  public DataFrame get(int[] obsNos, String[] name) throws Exception {
    DataFrame dframe = new DataFrame(obsNos.length);
    for(int i = 0; i < name.length; i++) {
      VariableInt var = (VariableInt) get(name[i]);
      if(var != null)
        dframe.add(name[i], (VariableInt) var.values(obsNos));
    }
    return(dframe);
  }

  public long numObservations() 
  { 
    return((long)length); 
  }

  public long numVariables() {
   return((long) size());
  }


  public org.omegahat.DataStructures.Data.VariableInt getVariable(String name) 
  {
   return((org.omegahat.DataStructures.Data.VariableInt)get(name));
  }


  public org.omegahat.DataStructures.Data.VariableInt addVariable(String name, org.omegahat.DataStructures.Data.VariableInt var) 
  {
   try {
     add(name, var);
   } catch(Exception ex){
     ex.printStackTrace();
   }
  //  put(name, var);
   return(var);
  }

  public boolean setNames(java.util.Collection els) {
   int curSize = size();
  int i = 0;
    for(java.util.Iterator e = els.iterator() ; e.hasNext() ; i++) {
      String name = e.next().toString();
       put(name, UnknownVariable);
    } 
    numHeadersSet += els.size();

  //System.err.println("Original size: " + curSize + " & number of headers read:  " + numHeadersSet);
   return(numHeadersSet >= curSize);
  }


  public String toString() {
    return("DataFrame: # obs " + numObservations() + ", # variables " + size());
  }

  /**
     Accessor for {@link #readHeader readHeader} field
   */
  public boolean getReadHeader()
  {
     return(readHeader);
  }

  /**
     Accessor for setting {@link #readHeader readHeader} field
   */

  public boolean setReadHeader(boolean value)
  {
     readHeader = value;
    return(readHeader);
  }


  public String[] rowNames()
  {
     return(rowNames);
  }

  public String[] rowNames(String[] value)
  {
     rowNames = value;
    return(rowNames);
  }

  public String[] rowNames(java.util.Vector v)
  {
   int n = v.size();
   String[] ans = new String[n];
   for(int i = 0; i < n; i++) {
     ans[i] = (String)v.elementAt(i);
   }
    return(rowNames(ans));
  }


  /**
     Accessor for {@link #metaData metaData} field
   */
  public java.util.Hashtable metaData()
  {
     return(metaData);
  }

  /**
     Accessor for setting {@link #metaData metaData} field
   */

  public java.util.Hashtable metaData(java.util.Hashtable value)
  {
     metaData = value;
    return(metaData);
  }




  /**
   This creates a {@link VariableInt VariableInt} object
   of the appropriate class  based on the type of 
   the argument <code>obj</code>.
   In this implementation, if this is a number, we create
   a {@link RealVariable  RealVariable}, and if it is a 
   {@link java.lang.String String} we create a 
   {@link FactorVariable FactorVariable}.
  */
  protected VariableInt createVariable(Object obj) {
   VariableInt var = null;
    if(obj instanceof String) {
      var = new FactorVariable();
  //    var = new FactorVariable((String)obj);
    } else if(obj instanceof Number) {
      var = new RealVariable();
  //((Number)obj).doubleValue()
    }

   return(var);
  }

  /**
   Create generic {@link RealVariable RealVariable} objects 
   and give each the name <code>Var<i>i</i></code>
   where <i>i</i> runs from 1 to <code>howMany</code>
  */
  protected int createVariables(int howMany) {
   for(int i = 0; i < howMany; i++)
      put("Var"+(i+1), createVariable(i));

   return(howMany);
  }

  public VariableInt createVariable(int which)
  {
    RealVariable var = new RealVariable();
    var.setName("Var"+which);

   return(var);
  }


  public VariableInt createDefaultVariable(String name, int num)
  {
   VariableInt var = new RealVariable(num);
    var.setName(name);
   return(var);
  }

  public VariableInt createDefaultVariable(String name)
  {
    return(createDefaultVariable(name, (int)numObservations()));
  }

  public void add(String name)
  {
    try {
      add(name, createDefaultVariable(name, (int)numObservations()));
    } catch(Exception ex) {
      System.err.println(ex);
    }
  }

  public String[] rowNames(java.io.InputStream stream)  
    throws java.io.IOException
  {
    java.util.Vector v = new java.util.Vector((int) numObservations());
    java.io.LineNumberReader reader = new java.io.LineNumberReader(new java.io.InputStreamReader(stream));
      try {
       String l;
        while((l = reader.readLine()) != null ) {
           l = l.trim();
           if(l.equals("") == false)
             v.add(l);
        }
      } catch(Throwable e) {
        e.printStackTrace();
      }

   return(rowNames(v));
  }

  public String rowName(String value, int which) {

   String[] tmp = rowNames();

    if(tmp == null) {
      tmp = rowNames(new String[(int)numObservations()]);
    }

    String old = null;
    old = tmp[which];

    tmp[which] = value;

  /*
   java.util.Vector v = rowNames();
    if(v == null)
     v = rowNames(new java.util.Vector((int)numObservations()));


   if(which < v.size()) {
     old = (String) v.elementAt(which);
     v.insertElementAt(value, which);
   }
   else 
    v.addElement(value);
  */

   return(old);
  }


  public org.omegahat.DataStructures.Data.BasicFrameInt subsetObservations(int[] whichRows)
  {
      throw new RuntimeException("Not implemented yet");    
  // return(this);
  }

  public org.omegahat.DataStructures.Data.BasicFrameInt subsetObservations(int start, int which)
  {
      throw new RuntimeException("Not implemented yet");
  // return(this);
  }



  /**
   Notification event called by a {@link org.omegahat.Environment.Tools.DataScanner.EventDataScanner EventDataScanner}
   when it has read a complete record.
  */
  public boolean newRecord(Object value, org.omegahat.Environment.Tools.DataScanner.ObjectReader source) {
   if(value instanceof java.util.Collection) {
     if(((java.util.Collection)value).size() == 0)
      return(false);

     if(getReadHeader()) {
        boolean done = setNames((java.util.Collection) value);
        if(done)
          setReadHeader(false);
        return(true);
     }

     add((java.util.Collection)value);
   }

   length++;

   return(true);
  } 

  /**
     Treat the entries in the collection as 
     components of a single record and append them
     to the corresponding variables, using the default
     ordering of the variable names
   @see #getVariableNames()
   @see #add(java.util.Iterator,String[])
  */
  public int add(java.util.Collection col)
  {
    if(getVariableNames() == null || getVariableNames().length == 0) {
      createVariables(col.size());
    }


   return( add(col.iterator()) );  
  }

  /**
   Process each element in the iterator as an element in a record by adding it
   to the variable identified by 
   the corresponding variable identified using the default names.
   @see #getVariableNames()
   @see #add(java.util.Iterator,String[])
  */
  public int add(java.util.Iterator els)
  {
    return( add(els, getVariableNames()) );  
  }

  /**
   Process each element in the iterator by adding it
   to the variable identified by 
   the corresponding  element in name.
  */
  public int add(java.util.Iterator els, String[] names)
  {
    int i;
    for(i = 0; i < names.length; i++) {
      Object val = els.next();
      VariableInt var = (VariableInt) get(names[i]);
       if(var == UnknownVariable) {
          var = createVariable(val);
          var.setName(getVariableNames()[i]);
          put(getVariableNames()[i], var);
       }

      var.add(val,-1);
    }

   return(i); 
  }




  /**
   Reads the records/observations for the data frame
   one at a time by being notified by the scanner.
   That this is separated from the constructors allows one to populate
   the data frame with variables before reading the actual record values.

  */
  public long read(org.omegahat.Environment.Tools.DataScanner.EventDataScanner scanner, boolean readHeader) {

   try {
       /* Register to be notified when a record is read completely. */
     scanner.addListener(this, true);
     setReadHeader(true);
       /* Only notify of a flushed record, not the end of a record. */
     scanner.readRecords(false);
   } catch(Throwable ex) {
     ex.printStackTrace();
   } finally {
     scanner.addListener(this);
   }

   return(numObservations());
  }

}

/*
Copyright (c) 1998 1999, The Omega Project for Statistical Computing.
     All rights reserved.

     Redistribution and use in source and binary forms, with or without
     modification, are permitted provided that the following conditions are
     met:

       Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

       Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

       Neither name of the Omega Project for Statistical Computing nor the 
       names of its contributors may be used to endorse or promote products 
       derived from this software without specific prior written permission.

     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY
     EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
     SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
     PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
     INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
     STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
     THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

*/

